import math
import os
import random
import time
# pip3 install pynput
# pip install pynput
from typing import List

from pynput import keyboard
from pynput.keyboard import Key

SCENE_WIDTH = 11
SCENE_HEIGHT = 15


class Position2D(object):
    def __init__(self, val_x, val_y, is_check_scene=True):
        super().__init__()
        self.__is_check_scene = is_check_scene
        self.__x = val_x
        self.__y = val_y

    @property
    def x(self): # get_x
        return self.__x

    @x.setter
    def x(self, val): # set_x
        if not self.__is_check_scene or 0 <= val < SCENE_WIDTH:
            self.__x = val

    @property
    def y(self): # get_y
        return self.__y

    @y.setter
    def y(self, val):  # set_y
        if not self.__is_check_scene or 0 <= val < SCENE_HEIGHT:
            self.__y = val

class Vector2D(Position2D):
    def __init__(self, val_x, val_y):
        super().__init__(val_x, val_y, is_check_scene=False)


class Element(object):
    def __init__(self): # constructor
        super().__init__()
        self._position = Position2D(0, 0)
        self._char = '⬜️'

    @property
    def char(self):
        return self._char

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, pos):
        self._position = pos

    def draw(self, scene):
        pos_x = int(self.position.x)
        pos_y = int(self.position.y)
        scene[pos_y][pos_x] = self.char

    def update(self, delta_time):
        pass

    def check_collsion(self, other):
        is_collision = False
        if int(other.position.x) == int(self.position.x) and \
                int(other.position.y) == int(self.position.y):
            is_collision = True
        return is_collision


class Wall(Element):
    def __init__(self, position): # constructor
        super().__init__()
        self._char = '🏛'
        self._position = position



class Explosion(Element):
    def __init__(self, position): # constructor
        super().__init__()
        self._char = '💥'
        self._position = position
        self._life = 0.5

    def update(self, delta_time):
        self._life -= delta_time
        if self._life < 0:
            GameState.instance().elements.remove(self)


class MovableElement(Element):
    def __init__(self): # constructor
        super().__init__()
        self._char = '🛸'
        self._speed = 1.0
        self._direction = Position2D(0, 0, is_check_scene=False)

    def update(self, delta_time):
        self.position.x += self._direction.x * self._speed * delta_time
        self.position.y += self._direction.y * self._speed * delta_time

    def stop(self):
        self._direction.x = 0
        self._direction.y = 0

    def left(self):
        self._direction.x = -1
        self._direction.y = 0

    def right(self):
        self._direction.x = 1
        self._direction.y = 0

    def up(self):
        self._direction.x = 0
        self._direction.y = -1

    def down(self):
        self._direction.x = 0
        self._direction.y = 1


class Player(MovableElement):
    def __init__(self):
        super().__init__()
        self._speed = 1.5
        self._char = '📤'

    def fire_rocket(self):
        rocket = Rocket(pos=Position2D(int(self.position.x), int(self.position.y)))
        rocket.up()
        GameState.instance().elements.append(rocket)
    
    def check_collsion(self, other):
        if type(other) != Rocket:
            super(Player, self).check_collsion(other)
        return is_collision

# TODO see changes
class Rocket(MovableElement):
    def __init__(self, pos: Position2D, is_up=True):
        super().__init__()
        self.position = pos
        self._speed = 1.5
        self._char = '🔺'
        if not is_up:
            self._char = '🔻'

    def update(self, delta_time):
        super().update(delta_time)
        if int(self.position.y) == 0:
            GameState.instance().elements.remove(self)


class Alien(MovableElement):
    def __init__(
            self,
            pos: Position2D,
            dir: Vector2D,
            listeners_aliens
    ):
        super().__init__()
        self.position = pos
        self._speed = 0.5
        self._char = '👾'
        self._direction = dir
        self._listeners_aliens = listeners_aliens # event listeners or observers
        self._listeners_aliens.append(self.notify) # not calling function but adding function as a pointer

        self._patience = 0
        self.reset_patience()

    def reset_patience(self):
        self._patience = 5 + random.random() * 10

    # TODO change to event based
    def notify(self, new_dir):
        self._direction = new_dir
        self._position.y += 1

    def check_border(self):
        is_at_border = False
        if round(self.position.x) == 0 or round(self.position.x) == SCENE_WIDTH:
            is_at_border = True
            self._direction.x *= -1
            for listener in self._listeners_aliens:
                listener(self._direction)
        return is_at_border

    # TODO fire_rocket

    # TODO update override

# TODO implement event based observers


class GameState(object): #singleton
    def __init__(self):
        super().__init__()
        if self._instance is not None:
            raise Exception('cannot init 2 singleton instances')
        self._instance = self

        pos_middle = Position2D(
            val_x=int(SCENE_WIDTH/2),
            val_y=SCENE_HEIGHT-1
        )
        self.player = Player()
        self.player.position = pos_middle
        self.elements: List = [
            self.player,
            Wall(position=Position2D(2, SCENE_HEIGHT-4)),
            Wall(position=Position2D(4, SCENE_HEIGHT-4)),
            Wall(position=Position2D(6, SCENE_HEIGHT-4)),
            Wall(position=Position2D(8, SCENE_HEIGHT-4))
        ]

        self.lives = 3
        self.score = 0

        self.listeners_aliens = []

        # TODO array of aliens 5x2 for loops

        rand_x = -1.0
        if random.random() > 0.5: # 0..1
            rand_x = 1.0
        alien = Alien(
            pos=Position2D(3, 3),
            dir=Vector2D(rand_x, 0.0),
            listeners_aliens=self.listeners_aliens
        )
        self.elements.append(alien)
        alien = Alien(
            pos=Position2D(4, 3),
            dir=Vector2D(rand_x, 0.0),
            listeners_aliens=self.listeners_aliens
        )
        self.elements.append(alien)
        alien = Alien(
            pos=Position2D(5, 3),
            dir=Vector2D(rand_x, 0.0),
            listeners_aliens=self.listeners_aliens
        )
        self.elements.append(alien)

        self.is_game_running = True

    _instance = None

    @staticmethod
    def instance():
        if GameState._instance is None:
            GameState._instance = GameState()
        return GameState._instance



elements = GameState.instance().elements
player = GameState.instance().player

# Get emojis:
# Windows Key + .
# Mac contrl + cmd + space

# main game loop
while GameState.instance().is_game_running:
    cmd_clear = 'clear'
    if os.name == 'nt':
        cmd_clear = 'cls'
    os.system(cmd_clear)

    scene = []
    for i in range(SCENE_HEIGHT): # 0...14
        columns = []
        for j in range(SCENE_WIDTH):
             columns.append('⬛️')
        scene.append(columns)

    for element in elements:
        element.draw(scene)

    scene_lines = []
    for line in scene:
        str_line = ''.join(line)
        scene_lines.append(str_line)
    str_scene = '\n'.join(scene_lines)
    print(f'Score: {GameState.instance().score}')
    print(str_scene)

    delay = 0.2 # seconds
    timestamp = time.time()

    with keyboard.Events() as events:
        key = events.get(delay)

        if key is not None: # key != null
            key_code = key.key
            if key_code == Key.left:
                player.left()
            elif key_code == Key.right: # else if
                player.right()
            elif key_code == Key.space:
                player.fire_rocket()
            elif key_code == Key.esc:
                GameState.instance().is_game_running = False
        else:
            player.stop()

    dt = delay - (time.time() - timestamp)
    if dt > 0:
        time.sleep(dt)

    for element in elements:
        if type(element) == Alien:
            element.check_border()

    dt = time.time() - timestamp
    for element in elements:
        element.update(dt)


    is_collision = False
    for i in range(len(elements)):
        for j in range(i+1, len(elements)):
            if (type(elements[i]) != Rocket and type(elements[j]) == Rocket) or \
                    (type(elements[j]) != Rocket and type(elements[i]) == Rocket):
                if elements[i].check_collsion(elements[j]):
                    pos = elements[i].position
                    del elements[j]
                    del elements[i]
                    elements.append(Explosion(pos))
                    is_collision = True
                	   break
        if is_collision:
            break
